/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: cosmicfits.h 161 2014-07-28 11:10:28Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2013 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *                         Arno Riffeser <arri@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#ifndef __COSMICFITS_H__
#define __COSMICFITS_H__

//#define LTL_RAMGE_CHECKING

//#define LTL_TEMPLATE_LOOP_LIMIT 64

#include "fitstools.h"

// next are already included within fitstools.h
//#include <cmath>

// #include <iostream>
//#include <list>

// #include <ltl/util/command_line_reader.h>
// #include <ltl/util/config_file_reader.h>
// #include <ltl/util/option_parser.h>

// #include <ltl/marray.h>
// #include <ltl/fitsio.h>
// #include <ltl/statistics.h>

#include <ltl/nonlinlsqfit.h>

#include "config.h.in"
#include "config.h"

// structure for flag values
class flagValues
{
   public:
      flagValues() : auto_adjust(false), region(2), verbose(1) { };
      double auto_adjust;
      double cosmic_fwhm;
      double circ_fwhm;
      double median_kappa;
      double replace_kappa;
      double confidence_kappa;
      double kappa;
      double start_fwhm;
      double start_angle;
      double circ_start;
      double alamda[4];
      int iter_max;
      int fit_radius;
      int replace_radius;
      int isolation;
      Region region; // region class defined in util/region.h
      float gain;
      float sky_offset;
      float readnoise;     
      float upper_error;
      float smooth_scale;
      float nan_value;
      float nan_error;
      bool refit_failed;
      bool mask;
      bool nan;
      string prefix;
      string usekey;
      string useerror;
      string fitskey;
      string fitsfilekey;
      bool zero;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

template<class TPAR, int NPAR, int NDIM>
class FitResult
{
   public:
      FixedVector<int, NDIM> coord;
      FVector<TPAR, NPAR> parameter;
      FVector<TPAR, NPAR> paramerror;
      TPAR chisquare;
      int niter;
};


void add_options(OptionParser& flags, flagValues& flagVal);

void print_usage(OptionParser& flags);

template<class T, int N>
void setMAborder(const T value, MArray<T, N>& a, const size_t width);


template<typename _RandomAccessIterator1, typename _RandomAccessIterator2>
size_t searchlocmax(const _RandomAccessIterator1 srcbegin,
                    const _RandomAccessIterator2 dstbegin,
                    const size_t xlength, const size_t ylength);

template<typename _RandomAccessIterator1,
         typename _RandomAccessIterator2,
         typename _NaNtype>
size_t masknanneighbors(const _RandomAccessIterator1 srcbegin,
                        const _RandomAccessIterator2 dstbegin,
                        const size_t xlength, const size_t ylength,
                        const _NaNtype nan,
                        const size_t limit);

// set width wide border of an MArray to value
template<class T, int N>
void setMAborder(const T value, MArray<T, N>& a, const size_t width)
{
   throw RangeException("specialisation of setMAborder missing");
}

template<class T>
void setMAborder(const T value, MArray<T, 2>& a, const size_t width)
{
   a( Range::all(),
      Range(a.minIndex(2), a.minIndex(2) + width - 1) ) = value;
   a( Range::all(),
      Range(a.maxIndex(2) - (width - 1), a.maxIndex(2)) ) = value;
   a( Range(a.minIndex(1), a.minIndex(1) + width - 1),
      Range::all() ) = value;
   a( Range(a.maxIndex(1) - (width - 1), a.maxIndex(1)),
      Range::all()) = value;
}

template<class T, int N>
void setMAoutsideRegion(const T value, MArray<T, N>& a,
                        const Region& region)
{
   throw RangeException("specialisation of setMAoutsideRegion missing");
}

template<class T>
void setMAoutsideRegion(const T value, MArray<T, 2>& a,
                        const Region& region)
{
   if(region.getStart(1) > a.minIndex(1))
      a( Range(a.minIndex(1), region.getStart(1) - 1), Range::all() ) = value;
   if(region.getEnd(1) < a.maxIndex(1))
      a( Range(region.getEnd(1) + 1, a.maxIndex(1)), Range::all() ) = value;
   if(region.getStart(2) > a.minIndex(2))
      a( Range::all(), Range(a.minIndex(2), region.getStart(2) - 1) ) = value;
   if(region.getEnd(2) < a.maxIndex(2))
      a( Range::all(), Range(region.getEnd(2) + 1, a.maxIndex(2)) ) = value;
}

// search for local maxima in 2D array
// ignore 1 pixel border
// check only on pixels where mask != 0
// modify mask by setting non maxima to 0
template<typename _RandomAccessIterator1, typename _RandomAccessIterator2>
size_t searchlocmax(const _RandomAccessIterator1 srcbegin,
                    const _RandomAccessIterator2 dstbegin,
                    const size_t xlength, const size_t ylength)
{
   const off_t xlenp1 = xlength + 1;
   const off_t xlenm1 = xlength - 1;
   size_t hitcounter = size_t(0);
   _RandomAccessIterator1 src_i;
   _RandomAccessIterator2 dst_i;
  
   for(off_t y = 1; y < off_t(ylength) - 1; ++y){
      const off_t pixeloffset  = 1 + y*xlength;
      dst_i = dstbegin + pixeloffset;
      src_i = srcbegin + pixeloffset;
      //cerr << "line: " << y << endl;
    
      for(off_t x = 1; x < xlenm1; ++x)
      {
         if(*dst_i){
            if(*src_i > *(src_i + 1)) {
               if(*src_i > *(src_i +  xlenm1)) {
                  if(*src_i > *(src_i + xlength)) {
                     if(*src_i > *(src_i + xlenp1)) {
                        if(*src_i > *(src_i - xlenp1)) {
                           if(*src_i > *(src_i - xlength)) {
                              if(*src_i > *(src_i - xlenm1)) {
                                 if(*src_i > *(src_i - 1)) {
                                    *(dst_i - 1) = 0;
                                    ++hitcounter;
                                 }
                                 else *dst_i = 0;
                                 *(dst_i - xlenm1) = 0;}
                              else *dst_i = 0;
                              *(dst_i - xlength) = 0;}
                           else *dst_i = 0;
                           *(dst_i - xlenp1) = 0;}
                        else *dst_i = 0;
                        *(dst_i + xlenp1) = 0;}
                     else *dst_i = 0;
                     *(dst_i + xlength) = 0;}
                  else *dst_i = 0;
                  *(dst_i + xlenm1) = 0;}
               else *dst_i = 0;
               *(dst_i + 1) = 0; ++dst_i; ++src_i; ++x;
            }
            else *dst_i = 0;
         }
         ++dst_i;
         ++src_i;
      }
   }

   return hitcounter;
}

// clear dst if more than limit neighbors in src equal nan
template<typename _RandomAccessIterator1,
         typename _RandomAccessIterator2,
         typename _NaNtype>
size_t masknanneighbors(const _RandomAccessIterator1 srcbegin,
                        const _RandomAccessIterator2 dstbegin,
                        const size_t xlength, const size_t ylength,
                        const _NaNtype nan,
                        const size_t limit)
{
   const off_t xlenp1 = xlength + 1;
   const off_t xlenm1 = xlength - 1;
   size_t hitcounter = size_t(0);
   _RandomAccessIterator1 src_i;
   _RandomAccessIterator2 dst_i;
   for(off_t y = 1; y < off_t(ylength) - 1; ++y){
      const off_t pixeloffset  = 1 + y*xlength;
      dst_i = dstbegin + pixeloffset;
      src_i = srcbegin + pixeloffset;
      for(off_t x = 1; x < xlenm1; ++x)
      {
         if(*dst_i){
            size_t invalid = 0;
            if(*(src_i - xlenp1)  == nan) ++invalid;
            if(*(src_i - xlength) == nan) ++invalid;
            if(*(src_i - xlenm1)  == nan) ++invalid;
            if(*(src_i - 1)       == nan) ++invalid;
            if(*(src_i + 1)       == nan) ++invalid;
            if(*(src_i +  xlenm1) == nan) ++invalid;
	    if(*(src_i + xlength) == nan) ++invalid;
            if(*(src_i + xlenp1)  == nan) ++invalid;
            if(invalid > limit)
            {
               *dst_i = 0;
               ++hitcounter;
            }
         }
         ++dst_i;
         ++src_i;
      }
   }
   return hitcounter;
}

// estimate a suitable maximum width for cosmicfwhm
template<class TPAR, int NPAR, int NDAT>
void estimate_cosmicfwhm(list< FitResult<TPAR, NPAR, NDAT> >& cosmics,
                         flagValues& flagVal)
{
   // get min fwhm histogram
   typename list< FitResult<TPAR, NPAR, NDAT> >::const_iterator cosmic_i =
      cosmics.begin();
   const typename list< FitResult<TPAR, NPAR, NDAT> >::const_iterator end =
      cosmics.end();
   MArray<TPAR, 1> minwidth(cosmics.size());
   typename MArray<TPAR, 1>::iterator min_i = minwidth.begin();
   while(cosmic_i != end)
   {
      FVector<TPAR, NPAR> par = (*cosmic_i).parameter;
      if( ( pow2(par(2) / flagVal.kappa) > 
            (((par(1) + flagVal.sky_offset) /
              flagVal.gain) + pow2(flagVal.readnoise)) ) // and high peak
         )
      {
         if(par(3) < par(4))
            *min_i = par(3);
         else *min_i = par(4);
      }
      else *min_i = -1.0;
      ++min_i;
      ++cosmic_i;
   }
   const size_t nobins = 200;
   const double startbin = 0.4;
   const double endbin = 20.4;
   const double step = (endbin - startbin) / double(nobins);
   const MArray<int, 1> hist( histogram( minwidth,
                                         nobins, startbin, endbin ) );
   // show histogram for debugging
   if(flagVal.verbose > 1)
   {
      typename MArray<int, 1>::const_iterator hist_i = hist.begin();
      int i = 0;
      while(!hist_i.done())
      {
         const double bin_i = startbin + step * double(i);
         if(flagVal.verbose > 1) {
            if(!(i%5)) cerr << endl; 
            cerr << bin_i << ": " << *hist_i << ", ";
         }
         ++hist_i;
         ++i;
      }
      cerr << endl;
   }

   // find first max, min, max sequence in histogram
   size_t i = 2;
   size_t max1 = 0, max2 = 0;
   double max1c = 0.0;
   double max2c = 0.0;

   double bin_i = 0.0;
   while( !max1 && (i < nobins) && (bin_i < 1.5))
   {
      bin_i = startbin + step * double(i-1);
      if( (hist(i) > hist(i-1)) && (hist(i) >= hist(i+1)) )
      {
         max1 = i;
         max1c = double(hist(i));
         if(max1c < flagVal.auto_adjust * sqrt(max1c))
         { max1 = 0; max1c = 0.0; }
      }
      ++i;
   }
   i = int((1.5 - startbin) / step) + 1;
   if(max1)
   {
      while( !max2 && (i < nobins) )
      {
         if( (hist(i) > hist(i-1)) && (hist(i) >= hist(i+1)) )
         {
            max2 = i;
            max2c = double(hist(i));
            if(max2c < flagVal.auto_adjust * sqrt(max2c))
            { max2 = 0; max2c = 0.0; }
         }
         ++i;
      }
   }

   double max1d(startbin + step * double(max1-1));
   double max2d(startbin + step * double(max2-1));
   if(max2d < 1.6)
   { max2 = 0; max2d = 0.0; }

   // check on significance
   const bool is_significant = (max1 && max2);
   if(is_significant)
      flagVal.cosmic_fwhm = (max1d + max2d) / 2.0;
   if(flagVal.verbose > 0)
   {
      cerr << "Cosmics width: " << max1d << "("<< max1c << ")"
           << ", PSF width: " << max2d  << "("<< max2c << ")"<< endl;
      if(is_significant) cerr << "S";
      else cerr << "Ins";
      cerr << "ignificant histogram at the " << flagVal.auto_adjust
           << " sigma confidence level:" << endl;
      if(is_significant)
         cerr << "Resetting cosmic_fwhm to ";
      else
         cerr << "Keeping cosmic_fwhm at ";
      cerr << flagVal.cosmic_fwhm << endl;
      cerr << "Still to use with caution!" << endl;
   }
}

// test if criteria for cosmic are matched
template<class TPAR, int NPAR, int NDAT>
inline bool is_cosmic(const FitResult<TPAR, NPAR, NDAT>& cosmic,
                      const flagValues& flagVal)
{
   // high peak
   const TPAR ampguess = cosmic.parameter(2);
   if(ampguess <= TPAR(0.0)) return false;
   const TPAR amperror = flagVal.confidence_kappa *
      sqrt(cosmic.paramerror(1) + cosmic.paramerror(2));
   const TPAR amplitude = (amperror >= ampguess) ?
      TPAR(0.0) :
      (ampguess - amperror);
   const TPAR noise2 =
      ((cosmic.parameter(1) + flagVal.sky_offset) / flagVal.gain) +
      pow2(flagVal.readnoise);
   if(noise2 <= TPAR(0.0)) return false;
   const bool is_low =
      ( pow2( amplitude / flagVal.kappa ) <= noise2 );
   if(is_low) return false;
   const TPAR xerror = flagVal.confidence_kappa * sqrt(cosmic.paramerror(3));
   const TPAR xwidth = cosmic.parameter(3) + xerror;
   // circular or elliptical?
   const bool is_circ = (cosmic.paramerror(4) == TPAR(0.0));
   // condition for circular
   if(is_circ)
      return ( xwidth <= flagVal.circ_fwhm );
   // condition for elliptical
   if( xwidth <= flagVal.cosmic_fwhm )
      return true;
   const TPAR yerror = flagVal.confidence_kappa * sqrt(cosmic.paramerror(4));
   const TPAR ywidth = cosmic.parameter(4) + yerror;
   return ( ywidth <= flagVal.cosmic_fwhm );
}

// replace cosmic in a and e_a according to flags, readjust mask for next pass
template<class TPAR, class TDAT, int NPAR, int NDAT>
inline size_t replace_cosmic(const FitResult<TPAR, NPAR, NDAT>& cosmic,
                             const flagValues& flagVal,
                             MArray<TDAT, NDAT>& e_b,
                             MArray<TDAT, NDAT>& a,
                             MArray<TDAT, NDAT>& e_a,
                             MArray<unsigned char, NDAT>& mask)
{
   // instance for gauss function
   PolyGaussian<TPAR, TDAT, NPAR, NDAT> gauss;
   // get ranges for replacement
   const int xmin = cosmic.coord(1) - flagVal.replace_radius;
   const int xstart = ((xmin > a.minIndex(1)) ? xmin : a.minIndex(1));
   const int xmax = cosmic.coord(1) + flagVal.replace_radius;
   const int ymin = cosmic.coord(2) - flagVal.replace_radius;
   const int ystart = ((ymin > a.minIndex(2)) ? ymin : a.minIndex(2));
   const int ymax = cosmic.coord(2) + flagVal.replace_radius;
   const Range xRange( xstart,
                       ((xmax < a.maxIndex(1)) ? xmax : a.maxIndex(1)) );
   const Range yRange( ystart,
                       ((ymax < a.maxIndex(2)) ? ymax : a.maxIndex(2)) );
   MArray<TDAT, NDAT> ra( a(xRange, yRange) );
   ra.setBase(xstart, ystart);
   MArray<TDAT, NDAT> e_ra( e_a(xRange, yRange) );
   e_ra.setBase(xstart, ystart);
   MArray<TDAT, NDAT> e_rb( e_b(xRange, yRange) );
   e_rb.setBase(xstart, ystart);
   MArray<TDAT, NDAT> rf( e_rb.shape() );
   rf.setBase(xstart, ystart);
   // fill in par(1) + par(2) * exp(...)
   // full version
//     gauss.fill( cosmic.parameter,
//                 cosmic.coord(1), cosmic.coord(2),
//                 rf );
//     const TPAR surf_noise2 =
//        ((cosmic.parameter(1) + flagVal.sky_offset) / flagVal.gain) + pow2(flagVal.readnoise);
//     const IndexList<NDAT> pixellist =
//        where( ( pow2((rf - cosmic.parameter(1)) / flagVal.replace_kappa) >
//                 surf_noise2 ) );
   // optimized version
   
   gauss.fillExp( cosmic.parameter,
                  cosmic.coord(1), cosmic.coord(2),
                  rf);
   const TPAR ln_surfnoise_kappa =
      log( (sqrt((cosmic.parameter(1) + flagVal.sky_offset) / flagVal.gain) +
            flagVal.readnoise) *
           flagVal.replace_kappa );

   const IndexList<NDAT> pixellist =
      where( (ln_surfnoise_kappa < rf) && (e_ra > TPAR(0.0)) );

   if(pixellist.size() < 1)
   {
      if(flagVal.verbose > 2)
         cerr << "   ambiguous detection\n"
              << rf << endl
              << cosmic.parameter << " @ " 
              << cosmic.coord(1) << ", " << cosmic.coord(2) << endl;
      return 0;
   }
   if(flagVal.verbose > 2)
   {
      for(typename list<FixedVector<int, NDAT> >::const_iterator il_i =
             pixellist.begin();
          il_i != pixellist.end();
          ++il_i)
         cerr << " " << setw(6) << "r"
              << " " << setw(6) << (*il_i)(1)
              << " " << setw(6) << (*il_i)(2) << endl;
   }
   // masking is now done at end by comparing with original
   // if(flagVal.mask)
   // {
   //    ra[pixellist] = flagVal.nan_value;
   //    e_ra[pixellist] = flagVal.nan_error;
   // }
   // else
   // {
      ra[pixellist] = cosmic.parameter(1);
      const TDAT surf_error2 = 
         TDAT( (cosmic.parameter(1) / flagVal.gain) + cosmic.paramerror(1) );
      for(typename list<FixedVector<int, NDAT> >::const_iterator il_i =
             pixellist.begin();
          il_i != pixellist.end();
          ++il_i)
         e_ra((*il_i)(1), (*il_i)(2)) += surf_error2;
   // }
   // set mask for next pass
   const int xmmin = cosmic.coord(1) - ( flagVal.replace_radius + 1 );
   const int xmmax = cosmic.coord(1) + ( flagVal.replace_radius + 1 );
   const int ymmin = cosmic.coord(2) - ( flagVal.replace_radius + 1 );
   const int ymmax = cosmic.coord(2) + ( flagVal.replace_radius + 1 );
   const Range xmRange( ((xmmin > mask.minIndex(1)) ?
                         xmmin : mask.minIndex(1)),
                        ((xmmax < mask.maxIndex(1)) ?
                         xmmax : mask.maxIndex(1)) );
   const Range ymRange( ((ymmin > mask.minIndex(2)) ?
                         ymmin : mask.minIndex(2)),
                        ((ymmax < mask.maxIndex(2)) ?
                         ymmax : mask.maxIndex(2)) );
   mask(xmRange, ymRange) = 1;
   return pixellist.size();
}

#endif
